from typing import Optional, Union
from collections import OrderedDict

import pandas as pd
import numpy as np
import pickle as pkl

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)
from matplotlib import gridspec
from matplotlib.figure import Figure

from solcore.light_source import LightSource
from solcore.constants import q


class IVPlot(object):
    """ Class that creates a plot of IV data, using a sensible formatting for the axis.
    """

    def __init__(self, data: pd.DataFrame, light: bool = False, spectrum: Optional[LightSource] = None,
                 title: str = '') -> None:
        """ Creates the plot of IV data

        :param data: Data to plot as a Pandas dataframe
        """
        self.data = data
        self.fig, self.axes = plt.subplots(num=title, figsize=(8, 6))
        self.axes.margins(0)

        # This will store the solar cell parameters in the case of light IV
        self.outputs = OrderedDict()

        if light:
            self.plot_light_iv(data, spectrum)
        else:
            self.plot_dark_iv(data)

        self.axes.set_xlabel('Voltage (V)')
        self.axes.set_ylabel('Current (A/m$^2$)')
        self.axes.legend(frameon=False)
        plt.tight_layout()

        # A small pause is needed to actually draw the plot
        plt.draw()
        plt.pause(0.001)

    def plot_dark_iv(self, data: pd.DataFrame) -> None:
        """ Plots the data assuming it is dark IV data, ie. absolute values for the current and log scale.

        :param data: Data to plot as a Pandas dataframe
        :return: None
        """

        for name in data.columns.values:

            # Plotting tunnel junctions IV
            if 'J_TJ' in name:
                label = 'Isolated {}'.format(name.split('_')[1])
                self.axes.semilogy(data['V'], abs(data[name]), label=label)

            # Plotting junctions IV, when isolated and as part of the MJ structure
            elif 'V_MJ' in name:
                id = name.split('V_MJ')[-1]
                p = self.axes.semilogy(data[name], abs(data['J']), 'o--', fillstyle='none', label=name.split('_')[1])
                color = p[-1].get_color()

                new_name = 'J_' + id
                label = 'Isolated {}'.format(id)
                self.axes.semilogy(data['V'], abs(data[new_name]), color=color, label=label)

        # Finally the voltage drop due to series resistance and the total IV.
        self.axes.plot(data['V_Rs'], abs(data['J']), 'o--', fillstyle='none', label='Rs')
        self.axes.semilogy(data['V'], abs(data['J']), 'k', linewidth=4, label='Total IV')

    def plot_light_iv(self, data: pd.DataFrame, spectrum: LightSource) -> None:
        """ Plots the data assuming it is light IV data, ie. moving shifting the region of interest to the first quadrant. It also calculates the solar cell parameters and the cell efficiency if a spectrum is given.

        :param data: Data to plot as a Pandas dataframe
        :param spectrum: A LightSource object with the spectrum, needed to calculate the efficiency of the cell
        :return: None
        """

        # We move the IV in the first quadrant, considering where Isc and Voc are
        sj = 1 if np.interp(0, data['V'], data['J']) >= 0 else -1
        sv = 1 if np.interp(0, data['J'], data['V']) >= 0 else -1

        # We have to track the maximum values for the axes, too.
        max_j = max(-1, np.interp(0, data['V'], sj * data['J']))
        max_v = max(-1, np.interp(0, data['J'], sv * data['V']))

        # In the process, we get the solar cell parameters
        power = sv * data['V'] * sj * data['J']
        idx_max = pd.Series.idxmax(power)
        self.outputs['jsc'] = max_j
        self.outputs['voc'] = max_v
        self.outputs['jmpp'] = sj * data['J'][idx_max]
        self.outputs['vmpp'] = sv * data['V'][idx_max]
        self.outputs['mpp'] = self.outputs['jmpp'] * self.outputs['vmpp']
        self.outputs['ff'] = self.outputs['mpp'] / (self.outputs['jsc'] * self.outputs['voc'])

        if spectrum is not None:
            self.outputs['eta'] = self.outputs['mpp'] / spectrum.power_density

        for name in data.columns.values:

            # Plotting tunnel junctions IV. Actually, we need to review this as TJ have opposite polarity than
            # normal J and therefore should look more like Rs.
            if 'J_TJ' in name:
                sj = 1 if np.interp(0, data['V'], data[name]) >= 0 else -1
                sv = 1 if np.interp(0, data[name], data['V']) >= 0 else -1
                max_j = max(max_j, np.interp(0, data['V'], sj * data[name]))
                label = 'Isolated {}'.format(name.split('_')[1])
                self.axes.plot(sv * data['V'], sj * data[name], label=label)

            # Plotting junctions IV, when isolated and as part of the MJ structure
            elif 'V_MJ' in name:
                id = name.split('V_MJ')[-1]
                sj = 1 if np.interp(0, data[name], data['J']) >= 0 else -1
                sv = 1 if np.interp(0, data['J'], data[name]) >= 0 else -1
                p = self.axes.plot(sv * data[name], sj * data['J'], 'o--', fillstyle='none', label=name.split('_')[1])
                color = p[-1].get_color()

                new_name = 'J_' + id
                sj = 1 if np.interp(0, data['V'], data[new_name]) >= 0 else -1
                sv = 1 if np.interp(0, data[new_name], data['V']) >= 0 else -1
                max_j = max(max_j, np.interp(0, data['V'], sj * data[new_name]))
                label = 'Isolated {}'.format(id)
                self.axes.plot(sv * data['V'], sj * data[new_name], color=color, label=label)

        # Finally the voltage drop due to series resistance and the total IV.
        sj = 1 if np.interp(0, data['V'], data['J']) >= 0 else -1
        sv = 1 if np.interp(0, data['J'], data['V']) >= 0 else -1
        self.axes.plot(sv * data['V_Rs'], sj * data['J'], 'o--', fillstyle='none', label='Rs')
        self.axes.plot(sv * data['V'], sj * data['J'], 'k', linewidth=4, label='Total IV')

        # And adjust the limits of the axis to fit everything inside
        self.axes.set_xlim(right=1.1 * max_v)
        self.axes.set_ylim(bottom=0, top=1.1 * max_j)
        self.axes.axvline(x=0, ymin=0, ymax=1.1 * max_j, color='k', linestyle='--')


class QEPlot(object):
    """ Class that creates a plot of QE data, using a sensible formatting for the axis.
    """

    def __init__(self, data: pd.DataFrame, spectrum: Optional[LightSource] = None, title: str = '') -> None:
        """ Creates the plot of QE data

        :param data: Data to plot as a Pandas dataframe
        :param spectrum: Solar spectrum, only needed to calculate the currents
        """
        self.data = data
        self.fig, self.axes = plt.subplots(num=title, figsize=(8, 6))
        self.axes.margins(0)

        # This will store the Jsc calculated for each junction if a solar spectrum is given
        self.outputs = OrderedDict()

        self.plot_qe(data, spectrum)

        self.axes.set_xlabel('Wavelength (nm)')
        self.axes.set_ylabel('QE (-)')
        self.axes.legend(frameon=False)
        plt.tight_layout()

        # A small pause is needed to actually draw the plot
        plt.draw()
        plt.pause(0.001)

    def plot_qe(self, data: pd.DataFrame, spectrum: LightSource) -> None:
        """ Plots the QE data and the different components, if any. If a spectrum is provided, it also calculates the Jsc of each junction

        :param data: Data to plot as a Pandas dataframe
        :param spectrum: A LightSource object with the spectrum, needed to calculate the Jsc of each junction
        :return: None
        """
        wl = data['WL'] * 1e9  # Wavelength in nanometers
        color = 'red'  # Default color; not really use
        i = 0  # Counter
        markerstyle = ('o', 'v', '^', '<', '>', '8', 's', 'p', '*', 'h', 'H', 'D', 'd', 'P', 'X')
        colors = ('r', 'g', 'b', 'm', 'c')
        losses = ('srh', 'rad', 'aug', 'surf', 'surb')
        losses_done = False

        # To accumulate the total light absorbed
        total_A = np.zeros_like(data['WL'])

        # And then the reflection
        y1 = 1 - data['R']
        y2 = np.ones_like(data['R'])
        self.axes.fill_between(wl, y1, y2, color='yellow', alpha=0.5, linewidth=3, label='R')

        # Now we plot what happens with light absorbed
        for name in data.columns.values:
            if 'A' in name:
                total_A += data[name]
                p = self.axes.plot(wl, data[name], '--', linewidth=3, label=name)
                color = p[-1].get_color()
                i = 0
                y1 = data[name]
            elif 'EQE' in name and len(name) <= 5:
                self.axes.plot(wl, data[name], linewidth=3, color=color, label=name)
            elif any(loss in name for loss in losses):
                # For the PDD model we have losses. We plot them as shaded areas
                label = losses[i] if (not losses_done and i < 5) else None
                y2 = y1
                y1 = y1 - data[name]
                self.axes.plot(wl, y1, ':', color=color)
                self.axes.fill_between(wl, y1, y2, facecolor=colors[i], alpha=0.5, label=label)
                i += 1
                losses_done = losses_done or i > 4
            elif 'EQE' in name:
                # And for the DA model we have contributions to the QE. We plot them with different markers.
                self.axes.plot(wl, data[name], color=color, marker=markerstyle[i], fillstyle='none', markevery=5,
                               label=name)
                i += 1

            # If a spectrum is given, we calculate the current related to each of the QE
            if spectrum is not None and name[:4] == 'EQE':
                current = data[name].values * spectrum.spectrum(x=wl, output_units='photon_flux_per_nm')[1]
                self.outputs[name] = q * np.trapz(current, wl)

        self.axes.plot(wl, total_A, 'k--', linewidth=2, label='Total A')

        self.axes.set_xlim(left=min(wl), right=max(wl))
        self.axes.set_ylim(bottom=0, top=1)


class BandstructurePlot(object):
    """ Class that creates a plot of bandstructure data and other position-dependent parameters data, using a sensible formatting for the axis.
    """

    def __init__(self, data: pd.DataFrame, title: str = '') -> None:
        """ Launches the creation of the plots, which actually consist on several separated subplots. We create:

        - Energy plot: Ec, Ev, Efe, Efh
        - Carrier density plot: Nc, Nv, Nd, Na, n, p, ni, rho  <-- Maybe a bit too much. Consider removing Nc, Nv
        - G-R plot: G and the recombination mechanisms SRH, Rad and Aug

        :param data: Data to plot as a Pandas dataframe
        """
        self.data = data
        self.fig, (self.energy, self.carriers, self.gr) = plt.subplots(num=title, figsize=(8, 8), nrows=3, ncols=1,
                                                                       sharex='all')

        self.plot_energy(data)
        self.plot_carriers(data)
        self.plot_gr(data)

        plt.tight_layout()

        # A small pause is needed to actually draw the plot
        plt.draw()
        plt.pause(0.001)

    def plot_energy(self, data: pd.DataFrame) -> None:
        """

        :param data:
        :return:
        """
        energy_data = ('Ec', 'Ev', 'Efe', 'Efh')
        colors_and_styles = {'Ec': 'b-', 'Ev': 'r-', 'Efe': 'b--', 'Efh': 'r--'}
        i = 0

        for name in data.columns.values:
            if any(ed in name for ed in energy_data):
                x = data['x_{}'.format(name.split('_')[-1])] * 1e9
                marker = colors_and_styles[name.split('_')[0]]
                label = name.split('_')[0] if i < 4 else ''
                self.energy.plot(x, data[name], marker, label=label)
                i += 1

        self.energy.set_ylabel('Energy (eV)')
        self.energy.tick_params(labelbottom=False)
        self.energy.legend(frameon=False)

    def plot_carriers(self, data: pd.DataFrame) -> None:
        """

        :param data:
        :return:
        """
        carrier_data = ('Nd', 'Na', 'n_', 'p_')
        colors_and_styles = {'Nd': 'b--', 'Na': 'r--', 'n': 'b-', 'p': 'r-'}
        i = 0

        for name in data.columns.values:
            if any(ed in name for ed in carrier_data):
                x = data['x_{}'.format(name.split('_')[-1])] * 1e9
                marker = colors_and_styles[name.split('_')[0]]
                label = name.split('_')[0] if i < 4 else ''
                self.carriers.semilogy(x, data[name] * 1e-6, marker, label=label)
                i += 1

        self.carriers.set_ylabel('Carrier density,\nDoping (cm$^{-3}$)')
        self.carriers.tick_params(labelbottom=False)
        self.carriers.legend(frameon=False)

    def plot_gr(self, data: pd.DataFrame) -> None:
        """

        :param data:
        :return:
        """
        self.gr.margins(0)
        gr_data = ('G', 'Rsrh', 'Rrad', 'Raug')
        colors_and_styles = {'G': 'r-', 'Rsrh': 'b-', 'Rrad': 'c-', 'Raug': 'g-'}
        i = 0

        for name in data.columns.values:
            if any(ed in name for ed in gr_data) and 'GR' not in name:
                x = data['x_{}'.format(name.split('_')[-1])] * 1e9
                marker = colors_and_styles[name.split('_')[0]]
                label = name.split('_')[0] if i < 4 else ''
                self.gr.semilogy(x, data[name] * 1e-6, marker, label=label)
                i += 1

        self.gr.set_ylabel('Generation,\nRecombination\n(cm$^{-3}$s$^{-1}$)')
        self.gr.set_xlabel('Position (nm)')
        self.gr.legend(frameon=False)


def save_plot(plot: Union[IVPlot, QEPlot, BandstructurePlot], filename: str) -> None:
    """ Saves a plot as a pickle object

    :param plot:
    :param filename:
    :return:
    """
    with open(filename, 'wb') as f:  # should be 'wb' rather than 'w'
        pkl.dump(plot, f)


def load_plot(filename: str, plot: bool = False) -> Union[IVPlot, QEPlot, BandstructurePlot]:
    """ Loads a plot from a pickle object

    :param filename: The name of the file to load
    :param plot: whether the figure should be plotted or not. Default=False.
    :return: An object containing the plot, of class IVPlot, QEPlot or BandstructurePlot
    """
    with open(filename, 'rb') as f:  # should be 'wb' rather than 'w'
        out = pkl.load(f)

    # A small pause is needed to actually draw the plot
    if plot:
        out.fig.canvas.draw_idle()
        plt.pause(0.001)

    return out


if __name__ == '__main__':
    import os

    # # df = pd.read_csv('DA_2J_IV_light.csv')
    # df = pd.read_csv('DA_2J_IV_dark.csv')
    #
    # # sp = LightSource(source_type='standard', version='AM1.5g', x=df['WL'])
    #
    df = pd.read_csv(os.path.join(os.path.expanduser('~'), 'Desktop', 'DA_2J_PDD_IV_light.csv'))
    my_fig = IVPlot(df, light=True, title='DA_2J_PDD_IV_light')
    plt.show()

    # df = pd.read_csv(os.path.join(os.path.expanduser('~'), 'Desktop','DA_2J_PDD_QE.csv'))
    #
    # sp = LightSource(source_type='standard', version='AM1.5g', x=df['WL'])
    #
    # my_fig = QEPlot(df, spectrum=sp, title='DA_2J_PDD_QE')
    # plt.show()

    # df = pd.read_csv(os.path.join(os.path.expanduser('~'), 'Desktop', 'DA_2J_PDD_equilibrium.csv'))
    #
    # my_fig = BandstructurePlot(df, title='DA_2J_PDD_equilibrium')
    # plt.show()

    # filename = 'DA_2J_PDD_short_circuit.csv'
    # df = pd.read_csv(os.path.join(os.path.expanduser('~'), 'Desktop', filename))
    #
    # my_fig = BandstructurePlot(df, title=filename)
    # plt.show()



import tkinter as tk
from tkinter import ttk

from solcore import material
from solcore.material_data import calculate_mobility
from solcore.constants import q

from solcore.solar_cell_solver import rcwa_options, pdd_options, asc_options

# The default material is a GaAs at 298 K and provides a set of minimum properties that are used if they are not
# available for the material of interest and not alternative is given
# Default values MUST BE in SI units
DefaultMaterial = material("GaAs")(T=298)
default_layer_properties = {'band_gap': DefaultMaterial.band_gap,  # J
                            'electron_affinity': DefaultMaterial.electron_affinity,  # J
                            'eff_mass_electron': DefaultMaterial.eff_mass_electron,  # relative to m0
                            'eff_mass_hh_z': DefaultMaterial.eff_mass_hh_z,  # relative to m0
                            'eff_mass_lh_z': DefaultMaterial.eff_mass_lh_z,  # relative to m0
                            'electron_mobility': calculate_mobility("GaAs", 0, 1),  # m2 V-1 s-1
                            'hole_mobility': calculate_mobility("GaAs", 1, 1),  # m2 V-1 s-1
                            'ni': DefaultMaterial.ni,  # m-3
                            'Nc': DefaultMaterial.Nc,  # m-3
                            'Nv': DefaultMaterial.Nv,  # m-3
                            'electron_minority_lifetime': 3e-6,  # s
                            'hole_minority_lifetime': 2.5e-7,  # s
                            'relative_permittivity': DefaultMaterial.relative_permittivity,  # relative to epsilon0
                            'electron_auger_recombination': 1e-42,  # m6 s-1,
                            'hole_auger_recombination': 1e-42,  # m6 s-1
                            'radiative_recombination': DefaultMaterial.radiative_recombination,  # m3 s-1
                            'Nd': 1,  # m-3
                            'Na': 1,  # m3
                            'electron_diffusion_length': 10e-6,  # m
                            'hole_diffusion_length': 40e-6}  # m

default_junction_properties = {'Sn': 1e10,
                               'Sp': 1e10,
                               'R_shunt': 1e14,
                               'R_series': 0,
                               'A': 1,
                               'Eg': 1.42,
                               'n': 3.5,
                               'jsc': 0,
                               'j01': 1e-15,
                               'j02': 1e-8,
                               'n1': 1,
                               'n2': 2,
                               'pn': 1,
                               'R': 1e-16,
                               'j_peak': 1,
                               'v_peak': 0.1,
                               'j_valley': 0.01,
                               'v_valley': 0.5,
                               'prefactor': 10}

# Units in which the parameters are shown in the GUI
units = {'Sn': 'cm2 s-1',
         'Sp': 'cm2 s-1',
         'R_shunt': 'Ohm',
         'R_series': 'Ohm',
         'A': '-',
         'Eg': 'eV',
         'n': '-',
         'jsc': 'mA cm-2',
         'j01': 'mA cm-2',
         'j02': 'mA cm-2',
         'n1': '-',
         'n2': '-',
         'band_gap': 'eV',
         'electron_affinity': 'eV',
         'relative_permittivity': '-',
         'Na': 'cm-3',
         'Nd': 'cm-3',
         'electron_mobility': 'cm2 V-1 s-1',
         'hole_mobility': 'cm2 V-1 s-1',
         'eff_mass_hh_z': '-',
         'eff_mass_lh_z': '-',
         'eff_mass_electron': '-',
         'electron_minority_lifetime': 'ns',
         'hole_minority_lifetime': 'ns',
         'radiative_recombination': 'cm3 s-1',
         'electron_auger_recombination': 'cm6 s-1',
         'hole_auger_recombination': 'cm6 s-1',
         'electron_diffusion_length': 'nm',
         'hole_diffusion_length': 'nm',
         'pn': '-',
         'R': 'Ohm',
         'j_peak': 'mA cm-2',
         'v_peak': 'V',
         'j_valley': 'mA cm-2',
         'v_valley': 'V',
         'prefactor': 'V-1'}

# Conversion factor such that parameter_GUI(units) = parameter(SI) / conversion factor
conversion = {'Sn': 1e-4,
              'Sp': 1e-4,
              'R_shunt': 1,
              'R_series': 1,
              'A': 1,
              'Eg': 1,
              'n': 1,
              'jsc': 10,
              'j01': 10,
              'j02': 10,
              'n1': 1,
              'n2': 1,
              'band_gap': q,
              'electron_affinity': q,
              'relative_permittivity': 1,
              'Na': 1e6,
              'Nd': 1e6,
              'electron_mobility': 1e-4,
              'hole_mobility': 1e-4,
              'eff_mass_hh_z': 1,
              'eff_mass_lh_z': 1,
              'eff_mass_electron': 1,
              'electron_minority_lifetime': 1e-9,
              'hole_minority_lifetime': 1e-9,
              'radiative_recombination': 1e-6,
              'electron_auger_recombination': 1e-12,
              'hole_auger_recombination': 1e-12,
              'electron_diffusion_length': 1e-9,
              'hole_diffusion_length': 1e-9,
              'pn': 1,
              'R': 1,
              'j_peak': 0.1,
              'v_peak': 1,
              'j_valley': 0.1,
              'v_valley': 1,
              'prefactor': 1}

# Set of parameters required for each layer or junction
properties_junctions = {'Junction-PDD': ['Sn', 'Sp', 'R_shunt', 'R_series'],
                        'Junction-DA': ['Sn', 'Sp', 'R_shunt', 'R_series'],
                        'Junction-2D': ['j01', 'j02', 'n1', 'n2', 'jsc', 'R_shunt', 'R_series'],
                        'Junction-DB': ['A', 'Eg', 'n', 'R_shunt', 'R_series'],
                        'TJ-resistive': ['R', 'pn'],
                        'TJ-parametric': ['j_peak', 'v_peak', 'j_valley', 'v_valley', 'prefactor', 'j01', 'pn']}

properties_layers = {'Junction-PDD': ['band_gap', 'electron_affinity', 'relative_permittivity', 'Na', 'Nd',
                                      'electron_mobility', 'hole_mobility',
                                      'eff_mass_hh_z', 'eff_mass_lh_z', 'eff_mass_electron',
                                      'electron_minority_lifetime', 'hole_minority_lifetime',
                                      'radiative_recombination', 'electron_auger_recombination',
                                      'hole_auger_recombination'],
                     'Junction-DA': ['band_gap', 'relative_permittivity', 'Na', 'Nd',
                                     'electron_mobility', 'hole_mobility', 'eff_mass_hh_z', 'eff_mass_lh_z',
                                     'eff_mass_electron', 'electron_diffusion_length', 'hole_diffusion_length'],
                     'Junction-2D': [],
                     'Junction-DB': [],
                     'TJ-resistive': [],
                     'TJ-parametric': []}

global_settings = [('T', 298, 'K'),
                   ('T_ambient', 298, 'K')]

electrical_settings = [('voltages', '0, 1.2, 100', 'V, V, points'),
                       ('internal_voltages', '-6, 4, 1000', 'V, V, points'),
                       ('mpp', False, ''),
                       ('light_iv', False, ''),
                       ('radiative_coupling', False, ''),
                       ('position', None, 'nm, nm, points')]

optical_settings = [('optics_method', 'BL', ''),
                    ('wavelength', '300, 1800, 251', 'nm, nm, points'),
                    ('source_type', 'standard', ''),
                    ('version', 'AM1.5g', ''),
                    ('output_units', 'photon_flux_per_m', '')]


class popupWindow(tk.Toplevel):
    def __init__(self, master, text, value, units):
        super(popupWindow, self).__init__(master)

        self.new_value = tk.StringVar(value=value)

        frame = ttk.Frame(self, padding=(10, 10, 10, 10))
        frame.grid()
        ttk.Label(frame, text=text).grid(column=0, row=0, columnspan=2, sticky=tk.NSEW)
        ttk.Entry(frame, textvariable=self.new_value).grid(column=0, row=1, sticky=tk.NSEW)
        ttk.Label(frame, text=units).grid(column=1, row=1, sticky=tk.NSEW)
        ttk.Button(frame, text='Ok', command=self.cleanup).grid(column=0, row=2, sticky=tk.NSEW)
        ttk.Button(frame, text='Cancel', command=self.exit).grid(column=1, row=2, sticky=tk.NSEW)

    def cleanup(self):
        self.value = self.new_value.get()
        self.destroy()

    def exit(self):
        self.value = None
        self.destroy()


class SettingsWindow(tk.Toplevel):
    def __init__(self, master, settings, model):
        super(SettingsWindow, self).__init__(master)

        frame = ttk.Frame(self, padding=(10, 10, 10, 10))
        frame.grid()

        self.settings = settings
        self.model = model

        # Junction properties
        self.settings_list = ttk.Treeview(frame)
        self.settings_list.grid(column=0, row=0, columnspan=2, sticky=(tk.NSEW))

        self.settings_list['columns'] = ('value', 'units')
        self.settings_list.heading('#0', text='Variable', anchor='w')
        self.settings_list.column("#0", minwidth=0, width=180, stretch=tk.NO)
        self.settings_list.heading('value', text='Value', anchor='w')
        self.settings_list.column("value", minwidth=0, width=180, stretch=tk.NO)
        self.settings_list.heading('units', text='Units', anchor='w')
        self.settings_list.column("units", minwidth=0, width=180, stretch=tk.NO)
        self.settings_list.tag_configure('modified', foreground='red')

        # When selecting an item, we update the properties frame
        self.settings_list.bind('<Double-Button-1>', self.settings_popup)

        ttk.Button(frame, text='Ok', command=self.cleanup).grid(column=0, row=2, sticky=tk.NSEW)
        ttk.Button(frame, text='Cancel', command=self.exit).grid(column=1, row=2, sticky=tk.NSEW)

        if settings == 'Global':
            self.load_global_settings()

        elif settings == 'Electrical':
            self.load_electrical_settings()

        elif settings == 'Optical':
            self.load_optical_settings()
        else:
            self.exit()

    def load_global_settings(self):
        """

        :return:
        """
        for value in global_settings:
            if value[0] in self.model['Global'].keys():
                self.settings_list.insert('', 'end', text=value[0], values=(self.model['Global'][value[0]], value[2]),
                                          tags=('modified'))
            else:
                self.settings_list.insert('', 'end', text=value[0], values=(value[1], value[2]))

    def load_electrical_settings(self):
        """

        :return:
        """
        for value in electrical_settings:
            if value[0] in self.model['Electrical'].keys():
                self.settings_list.insert('', 'end', text=value[0],
                                          values=(self.model['Electrical'][value[0]], value[2]),
                                          tags=('modified'))
            else:
                self.settings_list.insert('', 'end', text=value[0], values=(value[1], value[2]))

        for key in pdd_options.keys():
            if key in self.model['Electrical'].keys():
                self.settings_list.insert('', 'end', text=key, values=(self.model['Electrical'][key], ''),
                                          tags=('modified'))
            else:
                self.settings_list.insert('', 'end', text=key, values=(pdd_options[key], ''))

        for key in asc_options.keys():
            if key in self.model['Electrical'].keys():
                self.settings_list.insert('', 'end', text=key, values=(self.model['Electrical'][key], ''),
                                          tags=('modified'))
            else:
                self.settings_list.insert('', 'end', text=key, values=(asc_options[key], ''))

    def load_optical_settings(self):
        """

        :return:
        """
        for value in optical_settings:
            if value[0] in self.model['Optical'].keys():
                self.settings_list.insert('', 'end', text=value[0], values=(self.model['Optical'][value[0]], value[2]),
                                          tags=('modified'))
            else:
                self.settings_list.insert('', 'end', text=value[0], values=(value[1], value[2]))

        for key in rcwa_options.keys():
            if key in self.model['Optical'].keys():
                self.settings_list.insert('', 'end', text=key, values=(self.model['Optical'][key], ''),
                                          tags=('modified'))
            else:
                self.settings_list.insert('', 'end', text=key, values=(rcwa_options[key], ''))

    def settings_popup(self, *args):
        """

        :param args:
        :return:
        """
        item_id = self.settings_list.focus()
        text = self.settings_list.item(item_id, "text")
        value = self.settings_list.set(item_id, "value")
        units = self.settings_list.set(item_id, "units")

        w = popupWindow(self.master, 'New value for: ' + text, value, units)
        w.grab_set()
        self.master.wait_window(w)

        if w.value is not None and w.value != value:
            self.settings_list.set(item_id, "value", w.value)
            self.settings_list.item(item_id, tags=('modified'))

    def cleanup(self):
        """

        :return:
        """

        tagged = self.settings_list.tag_has('modified')

        for item_id in tagged:
            text = self.settings_list.item(item_id, "text")
            value = self.settings_list.set(item_id, "value")

            self.model[self.settings][text] = value

        self.destroy()

    def exit(self):
        """

        :return:
        """
        self.destroy()



import tkinter as tk
from tkinter import ttk

class LogTab(ttk.Frame):
    """ This tab is used to print the output log of any running process
    """
    def __init__(self, parent):
        """ Constructor of the class

        :param parent: The notebook that serves as parent of this tab
        """

        super(LogTab, self).__init__(parent)

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.text = tk.Text(self, wrap="word")
        scroll = ttk.Scrollbar(self, orient=tk.VERTICAL, command=self.text.yview)
        self.text.grid(column=0, row=0, sticky=(tk.NSEW))
        self.text.configure(yscrollcommand=scroll.set)
        scroll.grid(column=99, row=0, sticky=(tk.NS))

        self.text.tag_configure('red', foreground='red')



import tkinter as tk
from tkinter import ttk

from matplotlib import axes, figure
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)

import numpy as np

from solcore.light_source.light_source import LightSource, REGISTERED_CONVERTERS
from solcore import eVnm, nmHz, nmJ, mJ
from solcore.constants import q, h, c

from typing import Callable, Dict, Union


class SpectrumTab(ttk.Frame):
    """ This tab is used to create illumination spectra
    """

    def __init__(self, parent: ttk.Notebook) -> None:
        """ Constructor of the class

        :param parent: The notebook that serves as parent of this tab
        """
        super(SpectrumTab, self).__init__(parent)

        self.columnconfigure(1, weight=1)
        self.rowconfigure(0, weight=1)

        # We create the variables containing the details of the spectrum
        self.source_type = tk.StringVar(value='standard')
        self.concentration = tk.DoubleVar(value=1.0)
        self.units = tk.StringVar(value='power_density_per_nm')
        self.units_x = tk.StringVar(value='nm')
        self.x_min = tk.DoubleVar(value=300.0)
        self.x_max = tk.DoubleVar(value=1100.0)
        self.x_points = tk.IntVar(value=200)
        self.old_units = 'power_density_per_nm'
        self.old_units_x = 'nm'

        # We add some callbacks to update the plot when the range entries are changed
        self.x_min.trace_add("write", self.update_plot)
        self.x_max.trace_add("write", self.update_plot)
        self.x_points.trace_add("write", self.update_plot)

        # Next we create the variable that will hold the currently calculated LightSource object
        self.light_source = None

        # We add two frames, one for the controls on the left and another one for the plot on the right
        self.fig, self.axes = self.create_plot_area()
        self.control_frame = self.create_control_area()

        # We initialize the tab with the Standard AM1.5g solar spectrum
        self.spectrum = StandardSpectrum(parent=self.control_frame, row=4, update=self.update_plot)
        self.update_plot()

    def create_plot_area(self) -> (figure.Figure, axes.Axes):
        """ Creates the plot area in which to show the spectra

        :return: The figure and axes
        """
        plot_frame = ttk.Frame(self)
        plot_frame.grid(column=1, row=0, sticky=tk.NSEW)
        plot_frame.columnconfigure(0, weight=1)
        plot_frame.rowconfigure(0, weight=1)

        fig = figure.Figure(tight_layout=True)
        ax = fig.add_subplot(111)

        canvas = FigureCanvasTkAgg(fig, master=plot_frame)  # A tk.DrawingArea.
        canvas.draw()
        canvas.get_tk_widget().grid(column=0, row=0, sticky=tk.NSEW)

        toolbar_frame = ttk.Frame(master=plot_frame)
        toolbar_frame.grid(column=0, row=1, sticky=tk.NSEW)
        toolbar = NavigationToolbar2Tk(canvas, toolbar_frame)
        toolbar.update()

        return fig, ax

    def create_control_area(self):
        """

        :return:
        """
        control_frame = ttk.Frame(self)
        control_frame.grid(column=0, row=0, sticky=tk.NSEW)

        general_properties_frame = ttk.Frame(control_frame)
        general_properties_frame.grid(column=0, row=0, sticky=tk.NSEW)

        # Source type
        source_type_label = ttk.Label(general_properties_frame, text="Source type")
        spectrum_type_box = ttk.Combobox(general_properties_frame, state='readonly',
                                         values=LightSource.type_of_source, textvariable=self.source_type)
        source_type_label.grid(column=0, row=0, sticky=tk.NSEW)
        spectrum_type_box.grid(column=1, row=0, columnspan=2, sticky=tk.NSEW)
        spectrum_type_box.bind('<<ComboboxSelected>>', self.update_plot)

        # Units
        units_label = ttk.Label(general_properties_frame, text="Units")
        units_box = ttk.Combobox(general_properties_frame, state='readonly', values=list(REGISTERED_CONVERTERS.keys()),
                                 textvariable=self.units)
        units_label.grid(column=0, row=1, sticky=tk.NSEW)
        units_box.grid(column=1, row=1, columnspan=2, sticky=tk.NSEW)
        units_box.bind('<<ComboboxSelected>>', self.update_plot)

        # Range
        range_label = ttk.Label(general_properties_frame, text="Range (min, max, points) in: ")
        x_units_label = ttk.Label(general_properties_frame, textvar=self.units_x)
        x_max_label = ttk.Entry(general_properties_frame, textvar=self.x_min, width=10)
        x_min_label = ttk.Entry(general_properties_frame, textvar=self.x_max, width=10)
        x_points_label = ttk.Entry(general_properties_frame, textvar=self.x_points, width=10)
        range_label.grid(column=0, row=2, columnspan=2, sticky=tk.NSEW)
        x_units_label.grid(column=2, row=2, sticky=tk.NSEW)
        x_max_label.grid(column=0, row=3, sticky=tk.NSEW)
        x_min_label.grid(column=1, row=3, sticky=tk.NSEW)
        x_points_label.grid(column=2, row=3, sticky=tk.NSEW)

        return control_frame

    def update_plot(self, *args) -> None:
        """ Updates the plot whenever one of the settings changes and also update the LightSource variable

        :return: None
        """
        self.update_units_labels_and_values()
        self.light_source = self.calculate_light_source()
        x_label, y_label = self.get_axes_labels()

        self.axes.clear()
        self.axes.plot(*self.light_source.spectrum(), 'r')
        self.axes.set_xlabel(x_label)
        self.axes.set_ylabel(y_label)

        # recompute the ax.dataLim and update ax.viewLim using the new dataLim
        self.axes.relim()
        self.axes.autoscale_view()

        # re-draw the canvas
        self.fig.canvas.draw_idle()

    def update_units_labels_and_values(self) -> None:
        """ Updates the x units labels and fields

        :return: None
        """
        # If x units haven't changed, we do nothing
        new_x = self.units.get().split('_')[-1]
        old_x = self.units_x.get()
        if new_x == old_x:
            return

        self.units_x.set(new_x)

        old_min = self.x_min.get()
        old_max = self.x_max.get()

        if all(t in ('nm', 'eV') for t in (old_x, new_x)):
            new_min = eVnm(old_min)
            new_max = eVnm(old_max)
        elif all(t in ('nm', 'J') for t in (old_x, new_x)):
            new_min = nmJ(old_min)
            new_max = nmJ(old_max)
        elif all(t in ('nm', 'm') for t in (old_x, new_x)):
            factor = 1e-9 if old_x == 'nm' else 1e9
            new_min = factor * old_min
            new_max = factor * old_max
        elif all(t in ('nm', 'hz') for t in (old_x, new_x)):
            new_min = nmHz(old_min)
            new_max = nmHz(old_max)
        elif all(t in ('m', 'J') for t in (old_x, new_x)):
            new_min = mJ(old_min)
            new_max = mJ(old_max)
        elif all(t in ('m', 'eV') for t in (old_x, new_x)):
            factor = h * c / q
            new_min = factor / old_min
            new_max = factor / old_max
        elif all(t in ('m', 'hz') for t in (old_x, new_x)):
            factor = c
            new_min = factor / old_min
            new_max = factor / old_max
        elif all(t in ('J', 'eV') for t in (old_x, new_x)):
            factor = q if old_x == 'eV' else 1 / q
            new_min = factor * old_min
            new_max = factor * old_max
        elif all(t in ('J', 'hz') for t in (old_x, new_x)):
            factor = 1 / h if old_x == 'J' else h
            new_min = factor * old_min
            new_max = factor * old_max
        else:
            # eV <-> hz
            factor = q / h if old_x == 'eV' else h / q
            new_min = factor * old_min
            new_max = factor * old_max

        # Now we have to check if maximum and minimum are in the correct order, reversing them, otherwise
        if new_min > new_max:
            new_min, new_max = new_max, new_min

        self.x_min.set(format(new_min, '.4'))
        self.x_max.set(format(new_max, '.4'))

    def calculate_light_source(self) -> LightSource:
        """ Calculate the lightsource based on the chosen options

        :return: None
        """
        # First we create the x values
        try:
            x_min = self.x_min.get()
            x_max = self.x_max.get()
            x_points = self.x_points.get()
        except Exception:
            # If any of the parameters causes problems, we return the current light source
            return self.light_source

        # Now we retrieve the options
        options = self.spectrum.get_options()

        # And now we are ready to create the spectrum
        light_source = LightSource(source_type=self.source_type.get(),
                                   x=np.linspace(x_min, x_max, x_points),
                                   output_units=self.units.get(),
                                   concentration=self.concentration.get(),
                                   **options)

        return light_source

    def get_axes_labels(self) -> (str, str):
        """ Gets the correct labels for the axis

        :return: A tuple with the x and y labels
        """
        units = self.units.get()

        # First, the x axes
        x_units = units.split('_')[-1]
        if x_units in ('nm', 'm'):
            x_label = 'Wavelength ({})'.format(x_units)
        elif x_units == 'hz':
            x_label = 'Frequency (hz)'
        else:
            x_label = 'Energy ({})'.format(x_units)

        # And now the y axes
        if units.split('_')[0] == 'power':
            y_label = 'Power density (W m$^{{-2}}$ {}$^{{-1}}$)'.format(x_units)
        else:
            y_label = 'Photon flux (photons m$^{{-2}}$ {}$^{{-1}}$)'.format(x_units)

        return x_label, y_label


class StandardSpectrum(ttk.LabelFrame):
    """

    """

    def __init__(self, parent: ttk.Frame, row: int, update: Callable[[], None], col: int = 0) -> None:
        """

        :param parent:
        """
        super(StandardSpectrum, self).__init__(parent, text='Standard spectrum')
        self.grid(column=col, row=row, sticky=tk.NSEW)

        self.version = tk.StringVar(value='AM1.5g')

        source_version_label = ttk.Label(self, text="Version")
        source_version_box = ttk.Combobox(self, state='readonly', values=['AM0', 'AM1.5g', 'AM1.5d'],
                                          textvariable=self.version)
        source_version_label.grid(column=0, row=0, sticky=tk.NSEW)
        source_version_box.grid(column=1, row=0, columnspan=2, sticky=tk.NSEW)
        source_version_box.bind('<<ComboboxSelected>>', update)

    def get_options(self) -> Dict:
        """ Returns a dictionary with the options of this type of spectrum

        :return: A ditcionary with the options
        """
        out = {'version': self.version.get()}
        return out



import numpy as np
import pandas as pd
from collections import OrderedDict


def save_iv_data(root_filename, output, light_iv=False):
    """

    :return:
    """
    suffix = 'light' if light_iv else 'dark'
    filename = '{}_{}_{}.csv'.format(root_filename, 'IV', suffix)

    out = OrderedDict({'J': -output.iv['IV'][1], 'V': output.iv['IV'][0]})
    v = output.iv['IV'][0]

    # We get the IV of the normal junctions when included in the MJ structure
    for j in range(len(output.iv['junction IV'])):
        new_j = output.junction_indices[j]
        out['V_MJ{}'.format(new_j)] = output.iv['junction IV'][j][0]

    # And the series resistance
    out['V_Rs'] = output.iv['Rseries IV'][0]

    # We get the IV of the normal junctions if they were isolated
    for j in output.junction_indices:
        out['J_{}'.format(j)] = output[j].iv(v)

    # And the tunnel junctions
    for j in output.tunnel_indices:
        out['J_TJ{}'.format(j)] = output[j].iv(v)

    # Save the data as CSV file
    df = pd.DataFrame(out)
    df.to_csv(filename)
    return df, filename


def save_qe_data(root_filename, output):
    """

    :return:
    """

    qe_data = OrderedDict({'WL': output.wavelength,
                           'R': output.reflected})

    for j in range(len(output)):

        # All layers and junctions have defined a fraction of absorbed light, except some TJ and 2D kind of
        # junctions. We try to get that, first.
        try:
            new_key = 'A_{}'.format(j)
            z = np.arange(0, output[j].width, 1e-11)
            absorbed_per_wl = np.trapz(output[j].absorbed(z), z, axis=0)
            qe_data[new_key] = absorbed_per_wl

        except KeyError:
            print('WARNING: Element {} has no absorption defined.'.format(j))

        # We try to get the QE. If we have a single layer or TJ, this will fail as they don't have QE
        try:
            for k in output[j].qe.keys():
                if k != 'WL':
                    new_key = '{}_{}'.format(k, j)
                    qe_data[new_key] = output[j].qe[k]
        except AttributeError:
            print('WARNING: No "qe" calculated for element {}.'.format(j))

    filename = '{}_{}.csv'.format(root_filename, 'QE')
    df = pd.DataFrame.from_dict(qe_data)
    df.to_csv(filename)
    return df, filename


def save_bandstructure_data(root_filename, output, task):
    """

    :return:
    """
    bs_data = OrderedDict({})
    tsk = '{}_data'.format(task)

    for j in output.junction_indices:

        # Only junctions have information on bandstructure, and not all the models, so we take that into account.
        try:
            data = output[j].__dict__[tsk]['Bandstructure']
            x = data['x']

            if task == 'equilibrium':
                # The properties and the bandstructure are mapped at different positions, so we interpolate
                for p in output[j].__dict__[tsk]['Properties'].keys():
                    old_x = output[j].__dict__[tsk]['Properties']['x']
                    if p != 'x':
                        data[p] = np.interp(x, old_x, output[j].__dict__[tsk]['Properties'][p])
        except KeyError as err:
            print('WARNING: Junction {} has no bandstructure information.'.format(j))
            continue

        # So far, so good: we continue getting the data
        data['x'] += output[j].offset

        new_data = OrderedDict({})
        # We change the key names to include the junction number
        for key in data:
            new_key = '{}_{}'.format(key, j)
            new_data[new_key] = data[key]

        bs_data.update(new_data)

    # If some data has been added to the dictionary, we save that data.
    if len(bs_data.keys()) > 0:
        filename = '{}_{}.csv'.format(root_filename, task)
        df = pd.DataFrame(dict([(k, pd.Series(v)) for k, v in bs_data.items()]))
        df.to_csv(filename, index=False)
        return df, filename



import tkinter as tk
from tkinter import ttk

import solcore


class MaterialsTab(ttk.Frame):
    """ Class that contains all widgets related to the managing of Solcore materials

    """
    def __init__(self, parent, master):
        """ Constructor of the class

        :param parent: The notebook that serves as parent of this tab
        """

        super(MaterialsTab, self).__init__(parent)

        self.master = master

        self.populate_materials_list()

    def populate_materials_list(self):
        """

        :return:
        """

        self.mat_list = tk.Listbox(self)
        self.mat_list.bind('<<ListboxSelect>>', self.load_material_properties)
        mat_list_scroll = ttk.Scrollbar(self, orient=tk.VERTICAL, command=self.mat_list.yview)
        self.mat_list.configure(yscrollcommand=mat_list_scroll.set)
        self.mat_list.grid(column=1, row=0, columnspan=3, sticky=(tk.NSEW))
        mat_list_scroll.grid(column=0, row=0, sticky=(tk.NS))

        self.mat_properties = ttk.Treeview(self)
        mat_properties_scroll = ttk.Scrollbar(self, orient=tk.VERTICAL, command=self.mat_properties.yview)
        self.mat_properties.configure(yscrollcommand=mat_properties_scroll.set)
        self.mat_properties.grid(column=5, row=0, columnspan=3, sticky=(tk.NSEW))
        mat_properties_scroll.grid(column=4, row=0, sticky=(tk.NS))

        self.mat_properties['columns'] = ('value')
        self.mat_properties.heading('#0', text='Property', anchor='w')
        self.mat_properties.heading('value', text='Value', anchor='w')
        self.mat_properties.bind('<ButtonRelease-1>', self.print_properties)

        self.mat_database = solcore.ParameterSystem().database

        self.valid_materials = sorted(self.mat_database.sections())
        self.valid_materials.remove('Immediate Calculables')
        self.valid_materials.remove('Final Calculables')

        for material in self.valid_materials:
            self.mat_list.insert(tk.END, material)

    def load_material_properties(self, *args):
        """

        :param args:
        :return:
        """

        if args[0] in self.valid_materials:
            # This option gets all the properties of a given material and returns it
            material = args[0]
            T = args[1]
            x = args[2] if len(args) == 3 else 1

            parameters = {}

            for option in self.mat_database.options(material):
                parameters[option] = solcore.get_parameter(material, option, T=T, x=x)

            print(parameters)
            return parameters

        elif len(self.mat_list.curselection()) != 1:
            return

        else:
            self.mat_properties.delete(*self.mat_properties.get_children())
            self.mat_selected = self.mat_list.curselection()[0]
            material = self.valid_materials[self.mat_selected]

            for option in self.mat_database.options(material):
                x = solcore.get_parameter(material, option, T=300)
                self.mat_properties.insert('', 'end', text=option, values=(x))

    def print_properties(self, *args):
        curItem = self.mat_properties.focus()
        print(self.mat_properties.item(curItem))

if __name__ == '__main__':

    valid_materials = sorted(solcore.ParameterSystem().database.sections())
    valid_materials.remove('Immediate Calculables')
    valid_materials.remove('Final Calculables')

    for mat in valid_materials:
        print(mat)
        for option in solcore.ParameterSystem().database.options(mat):
            x = solcore.ParameterSystem().database.get(mat, option)
            print('\t {}  \t{}'.format(option, x))


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sunglass']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.4.3,<2.0.0', 'solcore>=5.7.7,<6.0.0']

setup_kwargs = {
    'name': 'sunglass',
    'version': '0.1.1',
    'description': 'GUI for easily use Solcore',
    'long_description': '# Sunglass\n\n*"What you need to look at Solcore!"*\n\nSunglass is a simple - and not yet fully operational - graphical user interface for the\nsolar cells and semiconductors modelling framework Solcore. \n\n## Using Sunglass\n\n1. Install it from PyPI\n\n```bash\npip install sunglass\n```\n\n2. If you want PDD support in Solcore, re-install Solcore with that support:\n\n```bash\npip install --no-deps --force-reinstall --install-option="--with_pdd" solcore\n```\n\n3. Run it!\n\n```bash\npython -m sunglass\n```\n\n## Develop Sunglass\n\n1. Install poetry following the instructions for your OS: https://python-poetry.org/docs/\n2. `git clone` this repository.\n3. Navigate to thee root folder and install sunglass with `poetry install`\n4. If you want PDD support in Solcore, re-install Solcore with that support:\n\n```bash\npip install --no-deps --force-reinstall --install-option="--with_pdd" solcore\n```',
    'author': 'Diego Alonso Álvarez',
    'author_email': 'd.alonso-alvarez@imperial.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ImperialCollegeLondon/sunglass',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import copy
import numpy as np

from solcore import material
from solcore.solar_cell import SolarCell
from solcore.structure import Layer, Junction, TunnelJunction
from solcore.solar_cell_solver import solar_cell_solver

from .properties_and_settings import *


def create_reflectivity(filename):
    """

    :param filename:
    :return:
    """
    return None


def run_solar_cell_model(task, model):
    """

    :param model:
    :return: a Solar cell object
    """

    # First, we take care of the options
    # ----------------------------------
    options = copy.copy(model['Global'])
    options.update(model['Electrical'])
    options.update(model['Optical'])
    T = float(options['T'])

    # We have to validate the options and check they have the correct format and type
    array_based_options = ['voltages', 'internal_voltages', 'position', 'wavelength']
    float_options = ['T', 'T_ambient']
    bool_options = ['mpp', 'light_iv', 'radiative_coupling']

    try:
        for p in options:
            # For the options that should be arrays, we create the arrays. The wavelength need ot be converted to meters
            if p in array_based_options:
                c = options[p].split(', ')
                ini = float(c[0])
                end = float(c[1])
                num = int(c[2])
                options[p] = np.linspace(ini, end, num) * 1e-9 ** (p == 'wavelength')

            # For the options that need to be floats
            elif p in float_options:
                options[p] = float(options[p])

            elif p in bool_options:
                options[p] = options[p].lower() in ['true', 'yes', 't', 1]

    except Exception as err:
        print('ERROR parsing the solver options: Option format not recognised')
        print(err)
        raise

    # Now we create the layers and junctions to the solar cell object
    # ---------------------------------------------------------------
    sc_data = model['Solar cell']
    all_layers = []
    all_materials = []

    for i in sc_data['structure']:
        current_element = sc_data[i[0]]

        # First the individual layers
        if 'Layer' in current_element['type']:
            layer_properties = {}
            width = current_element['width'] * 1e-9

            # Set the composition and get the properties, converting them to the correct units
            for key in current_element['options']:
                if key in ['element', 'x']:
                    layer_properties[current_element['options']['element']] = current_element['options']['x']
                else:
                    layer_properties[key] = current_element['options'][key] * conversion[key]

            all_materials.append(material(current_element['material'])(T=T, **layer_properties))

            all_layers.append(Layer(width, all_materials[-1], role=current_element['name']))
            continue

        # Unless it is an individual layer, we have junctions
        properties = {}
        for p in properties_junctions[current_element['type']]:
            properties[p] = default_junction_properties[p]

        properties.update(**current_element['options'])
        kind = current_element['type'].split('-')[-1]
        if 'TJ' in current_element['type']:
            new_junction = TunnelJunction(name=current_element['name'], kind=kind, T=T, **properties)
        else:
            new_junction = Junction(name=current_element['name'], kind=kind, T=T, **properties)

        # Now we add the layers, if any
        for l in i[1:]:
            current_child = sc_data[l]

            width = current_child['width'] * 1e-9
            layer_properties = {}

            # Set the composition and get the properties, converting them to the correct units
            for key in current_child['options']:
                if key in ['element', 'x']:
                    layer_properties[current_child['options']['element']] = current_child['options']['x']
                else:
                    layer_properties[key] = current_child['options'][key] * conversion[key]

            all_materials.append(material(current_child['material'])(T=T, **layer_properties))
            new_junction.append(Layer(width, all_materials[-1], role=current_child['name']))

        all_layers.append(new_junction)

    # Now we have to create the solar cell
    # ------------------------------------
    reflectivity = create_reflectivity(sc_data['reflectivity']) if sc_data['reflectivity'] != '' else None

    try:
        substrate = material(sc_data['substrate'])(T=T, **{sc_data['element']: sc_data['composition']})
    except KeyError:
        substrate = material(sc_data['substrate'])(T=T)

    sc = SolarCell(layers=all_layers, name=sc_data['name'], substrate=substrate, T=T, cell_area=sc_data['size'],
                   shading=sc_data['shading'], R_series=sc_data['r_series'], reflectivity=reflectivity)

    # With all said and done, we can run the solver
    # ---------------------------------------------
    solar_cell_solver(sc, task, user_options=options)
    print('Done!')

    return sc



import tkinter as tk
from tkinter import ttk
from tkinter import filedialog, messagebox

import yaml, os, copy, glob, time

import numpy as np
import pandas as pd
from collections import OrderedDict
import matplotlib.pyplot as plt

import solcore
from solcore import material
from solcore.solar_cell import SolarCell
from solcore.structure import Layer, Junction, TunnelJunction
from solcore.solar_cell_solver import solar_cell_solver

from .properties_and_settings import *
from .run_calculation import run_solar_cell_model
from .save_data import save_bandstructure_data, save_iv_data, save_qe_data
from .plot import BandstructurePlot, IVPlot, QEPlot, save_plot, load_plot


class SolarCellsTab(ttk.Frame):
    """ Class that contains all widgets related to the creation and simulation of solar cells

    """

    def __init__(self, parent, master):
        """ Constructor of the class

        :param parent: The notebook that serves as parent of this tab
        """
        super(SolarCellsTab, self).__init__(parent)

        self.master = master

        self.create_buttons()
        self.create_solar_cell_list()
        self.create_layer_properties_frame()
        self.create_junction_properties_frame()
        self.create_solar_cell_properties_frame()
        self.create_output_frame()

        # The following dictionary contains all the information related to the simulation, except material properties
        # taken as the default values.
        self.model = {'Global': {'T': 298},
                      'Electrical': {},
                      'Optical': {},
                      'Solar cell': {}}

        self.current_filename = 'Untitled.sol'
        self.current_location = os.path.expanduser('~')
        self.wd_label_var = tk.StringVar(value=self.current_location)
        wd_label = ttk.Label(self, textvar=self.wd_label_var, relief='sunken')
        wd_label.grid(column=0, row=500, columnspan=500, sticky=tk.NSEW, pady=5, padx=5)

        self.unselect_item()

    def create_solar_cell_list(self):
        """

        :return: None
        """

        self.solar_cell_list = ttk.Treeview(self)
        solar_list_scroll = ttk.Scrollbar(self, orient=tk.VERTICAL, command=self.solar_cell_list.yview)
        self.solar_cell_list.configure(yscrollcommand=solar_list_scroll.set)
        self.solar_cell_list.grid(column=2, row=0, columnspan=3, rowspan=99, sticky=tk.NSEW)
        solar_list_scroll.grid(column=1, row=0, sticky=tk.NS)

        # To un-select items an item that was selected
        self.solar_cell_list.bind("<Button-1>", self.unselect_item)
        self.solar_cell_list.bind("<Escape>", self.unselect_item)

        # When selecting an item, we update the properties frame
        self.solar_cell_list.bind("<<TreeviewSelect>>", self.show_properties_frame)

        self.solar_cell_list['columns'] = ('type', 'width', 'material', 'options')
        self.solar_cell_list.heading('#0', text='Name/Role', anchor='w')
        self.solar_cell_list.column("#0", minwidth=0, width=150, stretch=tk.NO)
        self.solar_cell_list.heading('type', text='Type', anchor='w')
        self.solar_cell_list.column("type", minwidth=0, width=120, stretch=tk.NO)
        self.solar_cell_list.heading('width', text='width (nm)', anchor='w')
        self.solar_cell_list.column("width", minwidth=0, width=80, stretch=tk.NO)
        self.solar_cell_list.heading('material', text='material', anchor='w')
        self.solar_cell_list.column("material", minwidth=0, width=80, stretch=tk.NO)
        self.solar_cell_list.heading('options', text='options', anchor='w')
        self.solar_cell_list.column("options", minwidth=0, width=200, stretch=tk.YES)

    def unselect_item(self, *args):
        """ Un-select the item from the list

        :param args:
        :return:
        """
        self.solar_cell_list.selection_remove(self.solar_cell_list.focus())
        self.solar_cell_list.focus('')

        # If no item is selected, we have to disable the layer properties frame
        self.layer_properties_frame.grid_remove()
        self.sc_properties_frame.grid()

    def create_buttons(self):
        """

        :return: None
        """
        # Buttons for loading and saving data
        loadsave_frame = ttk.LabelFrame(self, text='Load/Save model')
        loadsave_frame.grid(column=0, row=0, sticky=tk.NSEW)
        loadsave_frame.columnconfigure(0, weight=1)

        save = ttk.Button(loadsave_frame, text='Save', command=self.save_model)
        load = ttk.Button(loadsave_frame, text='Load', command=self.load_model)
        clear = ttk.Button(loadsave_frame, text='Clear', command=self.clear_solar_cell)
        save.grid(column=0, row=0, sticky=tk.NSEW)
        load.grid(column=0, row=1, sticky=tk.NSEW)
        clear.grid(column=0, row=2, sticky=tk.NSEW)

        # Buttons for adding the structure
        structure_frame = ttk.LabelFrame(self, text='Structure')
        structure_frame.grid(column=0, row=1, sticky=tk.NSEW)

        layer = ttk.Button(structure_frame, text='Add layer', width=15, command=self.add_layer)
        junction = ttk.Button(structure_frame, text='Add junction', width=15, command=self.add_junction)
        up = ttk.Button(structure_frame, text='▲', width=2, command=self.move_up)
        out = ttk.Button(structure_frame, text='×', width=2, command=self.delete_line)
        dup = ttk.Button(structure_frame, text='⥀', width=2, command=self.duplicate)
        down = ttk.Button(structure_frame, text='▼', width=2, command=self.move_down)
        layer.grid(column=0, row=0, columnspan=4, sticky=tk.NSEW)
        junction.grid(column=0, row=1, columnspan=4, sticky=tk.NSEW)
        up.grid(column=0, row=3, sticky=tk.NSEW)
        out.grid(column=1, row=3, sticky=tk.NSEW)
        dup.grid(column=2, row=3, sticky=tk.NSEW)
        down.grid(column=3, row=3, sticky=tk.NSEW)

        # Buttons for setting the conditions
        settings_frame = ttk.LabelFrame(self, text='Settings')
        settings_frame.grid(column=0, row=2, sticky=(tk.NSEW))
        settings_frame.columnconfigure(0, weight=1)

        glob = ttk.Button(settings_frame, text='Global', command=lambda: self.settings_popup(settings='Global'))
        elect = ttk.Button(settings_frame, text='Electrical',
                           command=lambda: self.settings_popup(settings='Electrical'))
        opt = ttk.Button(settings_frame, text='Optical', command=lambda: self.settings_popup(settings='Optical'))
        glob.grid(column=0, row=0, columnspan=1, sticky=tk.NSEW)
        elect.grid(column=0, row=1, columnspan=1, sticky=tk.NSEW)
        opt.grid(column=0, row=2, columnspan=1, sticky=tk.NSEW)

        # Buttons for running the solver
        running_frame = ttk.LabelFrame(self, text='Run')
        running_frame.grid(column=0, row=3, sticky=(tk.NSEW))
        running_frame.columnconfigure(0, weight=1)

        self.run_var = tk.IntVar(value=2)
        iv = ttk.Radiobutton(running_frame, text="IV", variable=self.run_var, value=2)
        qe = ttk.Radiobutton(running_frame, text="QE", variable=self.run_var, value=3)
        equi = ttk.Radiobutton(running_frame, text="Equilibrium", variable=self.run_var, value=4)
        sc = ttk.Radiobutton(running_frame, text="Short Circuit", variable=self.run_var, value=5)
        iv.grid(column=0, row=1, sticky=tk.EW)
        qe.grid(column=0, row=2, sticky=tk.EW)
        equi.grid(column=0, row=3, sticky=tk.EW)
        sc.grid(column=0, row=4, sticky=tk.EW)

        run = ttk.Button(running_frame, text='Run', command=self.run)
        run.grid(column=0, row=5, sticky=tk.NSEW)

    def move_up(self):
        """ Moves up an item

        :return: None
        """
        item_id = self.solar_cell_list.focus()
        self.solar_cell_list.move(item_id, self.solar_cell_list.parent(item_id),
                                  self.solar_cell_list.index(item_id) - 1)

    def move_down(self):
        """ Moves down an item

        :return: None
        """

        item_id = self.solar_cell_list.focus()
        self.solar_cell_list.move(item_id, self.solar_cell_list.parent(item_id),
                                  self.solar_cell_list.index(item_id) + 1)

    def delete_line(self):
        """ Deletes an item and moves the focus to the previous one within the same parent. When there are no more previous items, the command is ignored.

        :return: None
        """
        try:
            item_id = self.solar_cell_list.focus()
            other_item = self.solar_cell_list.prev(item_id)
            self.solar_cell_list.delete(item_id)
            self.model['Solar cell'].pop(item_id)
            self.solar_cell_list.focus(other_item)
        except:
            return

    def duplicate(self):
        """ Duplicates an item. Currently, it only works well for single layers as it does not duplicate the contents of a junction or tunnel junction, if present.

        :return: None
        """
        item_id = self.solar_cell_list.focus()
        parent_id = self.solar_cell_list.parent(item_id)
        item = self.solar_cell_list.item(item_id)
        new = self.solar_cell_list.insert(parent_id, tk.END, text=item['text'], values=item['values'], open=True)
        self.model['Solar cell'][new] = copy.copy(self.model['Solar cell'][item_id])

        # And we duplicate the children, if any
        for c in self.solar_cell_list.get_children(item_id):
            child_item = self.solar_cell_list.item(c)
            new_child = self.solar_cell_list.insert(new, tk.END, text=child_item['text'],
                                                    values=copy.copy(child_item['values']))
            self.model['Solar cell'][new_child] = copy.copy(self.model['Solar cell'][c])

    def add_layer(self):
        """ Adds a layer to the structure, either in the root or within a junction. The layer is also added to the model dictionary

        :return: None
        """
        item_id = self.solar_cell_list.focus()
        if 'Layer' not in self.solar_cell_list.item(item_id)['values']:
            new = self.solar_cell_list.insert(item_id, tk.END, text='New layer',
                                              values=('Layer', 100.0, 'GaAs', ''))
            self.model['Solar cell'][new] = {'name': 'New layer', 'type': 'Layer', 'width': 100.0,
                                             'material': 'GaAs',
                                             'options': {}}

    def add_junction(self):
        """ Adds a junction to the structure

        :return: None
        """
        item_id = self.solar_cell_list.insert('', tk.END, text='New junction', values=('Junction-PDD', '', '', ''),
                                              open=True)
        self.model['Solar cell'][item_id] = {'name': 'New junction', 'type': 'Junction-PDD', 'width': '',
                                             'material': '',
                                             'options': {}}

    def clear_solar_cell(self, confirmed=False):
        """ Delets all elements in the slar cell

        :return:
        """

        clear = True
        if not confirmed:
            clear = messagebox.askyesno("Are you sure you want to delete current model?",
                                        "This action cannot be undone!",
                                        icon='warning')

        if clear:
            old_list = self.solar_cell_list.get_children()
            if len(old_list) > 0:
                self.solar_cell_list.delete(*old_list)

            self.model = {'Global': {'T': 298},
                          'Electrical': {},
                          'Optical': {},
                          'Solar cell': {}}

            self.current_filename = 'Untitled.sol'
            self.current_location = os.path.expanduser('~')

    def settings_popup(self, settings):
        """

        :return:
        """
        w = SettingsWindow(self.master, settings, self.model)
        w.grab_set()
        self.master.wait_window(w)

    def properties_popup(self, *args):
        """

        :return:
        """
        element_id = self.solar_cell_list.focus()

        if self.model['Solar cell'][element_id]['type'] == 'Layer':
            # We are editing the properties of a layer
            widget = self.properties_list
        else:
            # We are editing the properties of a junction
            widget = self.junction_properties_list

        item_id = widget.focus()
        text = widget.item(item_id, "text")
        value = widget.set(item_id, "value")
        units = widget.set(item_id, "units")

        w = popupWindow(self.master, 'New value for: ' + text, value, units)
        w.grab_set()
        self.master.wait_window(w)

        if w.value is not None and float(w.value) != value:
            widget.set(item_id, "value", float(w.value))
            widget.item(item_id, tags=('modified'))
            self.model['Solar cell'][element_id]['options'][text] = float(w.value)

    def create_solar_cell_properties_frame(self, *args):
        """ Creates the frame that contain the general solar cell properties, like name, substrate, front relfection...

        :param args:
        :return:
        """
        self.sc_properties_frame = ttk.LabelFrame(self, text='Solar cell properties', width=380)
        self.sc_properties_frame.grid(column=10, row=0, rowspan=30, sticky=tk.NSEW)
        self.sc_properties_frame.columnconfigure(1, weight=1)
        self.sc_properties_frame.grid_propagate(0)

        # Name of the cell
        ttk.Label(self.sc_properties_frame, text="Name:").grid(column=0, row=0, sticky=tk.NSEW)
        self.sc_name_var = tk.StringVar(value='New solar cell')
        sc_name = ttk.Entry(self.sc_properties_frame, textvariable=self.sc_name_var)
        sc_name.grid(column=1, row=0, sticky=tk.NSEW)

        # Substrate material
        ttk.Label(self.sc_properties_frame, text="Substrate:").grid(column=0, row=1, sticky=(tk.NSEW))
        self.substrate_var = tk.StringVar(value='GaAs')
        substrate_box = ttk.Combobox(self.sc_properties_frame, state='readonly',
                                     values=self.master.materials.valid_materials,
                                     textvariable=self.substrate_var)
        substrate_box.grid(column=1, row=1, sticky=tk.NSEW)
        substrate_box.bind('<<ComboboxSelected>>', self.update_substrate_composition_label)

        # Composition
        self.subs_composition_label_var = tk.StringVar(value='-')
        lab = ttk.Label(self.sc_properties_frame, textvariable=self.subs_composition_label_var)
        lab.grid(column=0, row=2, sticky=tk.NSEW)
        self.subs_composition_var = tk.DoubleVar(value=0)
        subs_composition = tk.Spinbox(master=self.sc_properties_frame, from_=0, to=1,
                                      textvariable=self.subs_composition_var)
        subs_composition.grid(column=1, row=2, sticky=tk.NSEW)

        # Reflectivity
        self.reflectivity_check = tk.IntVar(value=0)
        reflect = ttk.Checkbutton(self.sc_properties_frame, text='Reflectivity:', var=self.reflectivity_check,
                                  command=self.choose_reflectivity)
        reflect.grid(column=0, row=3, sticky=tk.NSEW)
        self.reflectivity_var = tk.StringVar(value='')
        reflectivity = ttk.Entry(self.sc_properties_frame, textvariable=self.reflectivity_var)
        reflectivity.grid(column=1, row=3, sticky=tk.NSEW)

        # Shading
        ttk.Label(self.sc_properties_frame, text="Shading").grid(column=0, row=4, sticky=tk.NSEW)
        self.shading_var = tk.DoubleVar(value=0)
        shading = tk.Spinbox(master=self.sc_properties_frame, from_=0, to=1, textvariable=self.shading_var)
        shading.grid(column=1, row=4, sticky=tk.NSEW)

        # Size
        ttk.Label(self.sc_properties_frame, text="Size (cm2):").grid(column=0, row=5, sticky=tk.NSEW)
        self.size_var = tk.DoubleVar(value=1)
        size = tk.Spinbox(self.sc_properties_frame, textvariable=self.size_var, from_=0.0001, to=1e6)
        size.grid(column=1, row=5, sticky=tk.NSEW)

        # Series resistance
        ttk.Label(self.sc_properties_frame, text="R series (Ohm):").grid(column=0, row=6, sticky=tk.NSEW)
        self.rs_var = tk.DoubleVar(value=0)
        rs = tk.Spinbox(self.sc_properties_frame, textvariable=self.rs_var, from_=0)
        rs.grid(column=1, row=6, sticky=tk.NSEW)

    def choose_reflectivity(self, *args):
        """ Allows to choose the reflectivity of the solar cell

        :param args:
        :return:
        """
        if self.reflectivity_check.get() == 0:
            return
        else:
            # A pop up dialog allows to choose the relfectivity
            filename = filedialog.askopenfilename(initialdir=self.current_location)

            if filename is not '':
                self.reflectivity_var.set(filename)
            else:
                self.reflectivity_check.set(0)

    def update_substrate_composition_label(self, *args):
        """ Updates the composition label of the substrate with the relevant material name

        :param args:
        :return:
        """
        # Checking if the material is ternary (or more) by counting capital letters.
        # If it is not, we ignore the compositionon
        mat = self.substrate_var.get()
        if sum(1 for c in mat if c.isupper()) > 2:
            x_label = solcore.ParameterSystem().database.get(mat, 'x')
            self.subs_composition_label_var.set(x_label)
        else:
            self.subs_composition_label_var.set('-')

    def confirm_changes_to_solar_cell(self, *args):
        """ Confirm all changes done to the solar cell properties by storing them in the model

        :param args:
        :return:
        """
        self.model['Solar cell']['name'] = self.sc_name_var.get()
        self.model['Solar cell']['substrate'] = self.substrate_var.get()
        self.model['Solar cell']['r_series'] = self.rs_var.get()
        self.model['Solar cell']['reflectivity'] = self.reflectivity_var.get()
        self.model['Solar cell']['shading'] = self.shading_var.get()
        self.model['Solar cell']['size'] = self.size_var.get()

        if self.subs_composition_label_var.get() != '-':
            self.model['Solar cell']['element'] = self.subs_composition_label_var.get()
            self.model['Solar cell']['composition'] = self.subs_composition_var.get()

    def create_junction_properties_frame(self, *args):
        """ Creates the frame that contain the junction properties

        :param args:
        :return:
        """
        self.junction_properties_frame = ttk.LabelFrame(self, text='Junction properties', width=380)
        self.junction_properties_frame.grid(column=10, row=0, rowspan=30, sticky=tk.NSEW)
        self.junction_properties_frame.columnconfigure(1, weight=1)
        self.junction_properties_frame.grid_propagate(0)

        # Name of the cell
        ttk.Label(self.junction_properties_frame, text="Name:").grid(column=0, row=0, sticky=tk.NSEW)
        self.junction_name_var = tk.StringVar(value='New junction')
        junction_name = ttk.Entry(self.junction_properties_frame, textvariable=self.junction_name_var)
        junction_name.grid(column=1, row=0, sticky=tk.NSEW)

        # Type of junction
        keys = [k for k in properties_junctions.keys()]
        ttk.Label(self.junction_properties_frame, text="Type:").grid(column=0, row=1, sticky=(tk.NSEW))
        self.junction_type_var = tk.StringVar(value='Junction-PDD')
        junction_type_box = ttk.Combobox(self.junction_properties_frame, state='readonly',
                                         values=keys,
                                         textvariable=self.junction_type_var)
        junction_type_box.grid(column=1, row=1, sticky=tk.NSEW)
        junction_type_box.bind('<<ComboboxSelected>>', self.update_junction_properties_list)

        # Junction properties
        self.junction_properties_list = ttk.Treeview(self.junction_properties_frame)
        self.junction_properties_list.grid(column=0, row=3, columnspan=2, sticky=(tk.NSEW))

        self.junction_properties_list['columns'] = ('value', 'units')
        self.junction_properties_list.heading('#0', text='Variable', anchor='w')
        self.junction_properties_list.column("#0", minwidth=0, width=180, stretch=tk.NO)
        self.junction_properties_list.heading('value', text='Value', anchor='w')
        self.junction_properties_list.column("value", minwidth=0, width=100, stretch=tk.NO)
        self.junction_properties_list.heading('units', text='Units', anchor='w')
        self.junction_properties_list.column("units", minwidth=0, width=100, stretch=tk.NO)
        self.junction_properties_list.tag_configure('modified', foreground='red')

        # When selecting an item, we update the properties frame
        self.junction_properties_list.bind('<Double-Button-1>', self.properties_popup)

        confirm = ttk.Button(self.junction_properties_frame, text='Confirm changes to junction',
                             command=self.confirm_changes_to_junction)
        confirm.grid(column=0, row=99, columnspan=2, sticky=tk.NSEW)

    def update_junction_properties_list(self, *args):
        """

        :return:
        """

        item_id = self.solar_cell_list.focus()

        self.model['Solar cell'][item_id]['type'] = self.junction_type_var.get()

        # We populate the properties TreeView with the values for this junction,
        # starting by eliminating the current ones
        old_list = self.junction_properties_list.get_children()
        if len(old_list) > 0:
            self.junction_properties_list.delete(*old_list)

        # Get the new set of parameters
        prop_list = self.get_properties_list(item_id)

        # We scan the list of required properties and load the overwritten ones of the default ones
        for key in prop_list:
            try:
                prop = self.model['Solar cell'][item_id]['options'][key]
                self.junction_properties_list.insert('', 'end', text=key, values=(format(prop, '.4'), units[key]),
                                                     tags=('modified'))
            except:
                prop = default_junction_properties[key] / conversion[key]

                self.junction_properties_list.insert('', 'end', text=key, values=(format(prop, '.4'), units[key]))

    def confirm_changes_to_junction(self, *args):
        """

        :param args:
        :return:
        """
        item_id = self.solar_cell_list.focus()

        junction_type = self.model['Solar cell'][item_id]['name'] = self.junction_type_var.get()
        name = self.model['Solar cell'][item_id]['name'] = self.junction_name_var.get()

        options = ''
        for key in self.model['Solar cell'][item_id]['options']:
            options += '{0}={1} '.format(key, self.model['Solar cell'][item_id]['options'][key])

        # We update the layers list
        self.solar_cell_list.item(item_id, text=name, values=[junction_type, '', '', options])

    def create_layer_properties_frame(self, *args):
        """

        :return: None
        """

        name = 'New layer'
        mat = 'GaAs'
        width = 100

        self.layer_properties_frame = ttk.LabelFrame(self, text='Layer properties', width=380)
        self.layer_properties_frame.grid(column=10, row=0, rowspan=30, sticky=tk.NSEW)
        self.layer_properties_frame.columnconfigure(1, weight=1)
        self.layer_properties_frame.rowconfigure(4, weight=1)
        self.layer_properties_frame.grid_propagate(0)

        ttk.Label(self.layer_properties_frame, text="Name/Role:").grid(column=0, row=0, sticky=(tk.NSEW))
        self.name_var = tk.StringVar(value=name)
        self.name = ttk.Entry(self.layer_properties_frame, textvariable=self.name_var)
        self.name.grid(column=1, row=0, sticky=(tk.NSEW))

        ttk.Label(self.layer_properties_frame, text="Width (nm):").grid(column=0, row=1, sticky=(tk.NSEW))
        self.width_var = tk.DoubleVar(value=width)
        self.width = tk.Spinbox(master=self.layer_properties_frame, from_=0, textvariable=self.width_var)
        self.width.grid(column=1, row=1, sticky=(tk.NSEW))

        ttk.Label(self.layer_properties_frame, text="Material:").grid(column=0, row=2, sticky=(tk.NSEW))
        self.selected_material = tk.StringVar(value=mat)
        self.materials_box = ttk.Combobox(self.layer_properties_frame, state='readonly',
                                          values=self.master.materials.valid_materials,
                                          textvariable=self.selected_material)
        self.materials_box.grid(column=1, row=2, sticky=(tk.NSEW))
        self.materials_box.bind('<<ComboboxSelected>>', self.update_composition_label)

        self.composition_label_var = tk.StringVar(value='-')
        ttk.Label(self.layer_properties_frame, textvariable=self.composition_label_var).grid(column=0, row=3,
                                                                                             sticky=(tk.NSEW))
        self.composition_var = tk.DoubleVar(value=0)
        self.composition = tk.Spinbox(master=self.layer_properties_frame, from_=0, to=1,
                                      textvariable=self.composition_var)
        self.composition.grid(column=1, row=3, sticky=(tk.NSEW))

        self.properties_list = ttk.Treeview(self.layer_properties_frame)
        self.properties_list.grid(column=0, row=4, columnspan=2, sticky=(tk.NSEW))

        self.properties_list['columns'] = ('value', 'units')
        self.properties_list.heading('#0', text='Variable', anchor='w')
        self.properties_list.column("#0", minwidth=0, width=180, stretch=tk.NO)
        self.properties_list.heading('value', text='Value', anchor='w')
        self.properties_list.column("value", minwidth=0, width=100, stretch=tk.NO)
        self.properties_list.heading('units', text='Units', anchor='w')
        self.properties_list.column("units", minwidth=0, width=100, stretch=tk.NO)
        self.properties_list.tag_configure('modified', foreground='red')

        # When selecting an item, we update the properties frame
        self.properties_list.bind('<Double-Button-1>', self.properties_popup)

        confirm = ttk.Button(self.layer_properties_frame, text='Confirm changes to layer',
                             command=self.confirm_changes_to_layer)
        confirm.grid(column=0, row=99, columnspan=2, sticky=tk.NSEW)

    def show_properties_frame(self, *args):
        """

        :return: None
        """

        item_id = self.solar_cell_list.focus()

        if item_id == '':
            return
        elif 'Layer' in self.model['Solar cell'][item_id]['type']:
            self.sc_properties_frame.grid_remove()
            self.junction_properties_frame.grid_remove()
            self.show_layer_properties_frame(item_id)
        elif 'Junction' in self.model['Solar cell'][item_id]['type']:
            self.sc_properties_frame.grid_remove()
            self.layer_properties_frame.grid_remove()
            self.show_junction_properties_frame(item_id)

    def show_junction_properties_frame(self, item_id):
        """

        :return: None
        """
        current_item = self.model['Solar cell'][item_id]

        # We bring the properties frame back, in case it was hidden
        self.junction_properties_frame.grid()

        self.junction_name_var.set(current_item['name'])
        self.junction_type_var.set(current_item['type'])

        # Finally, we populate the properties TreeView with the values for this junction,
        # starting by eliminating the current ones
        old_list = self.junction_properties_list.get_children()
        if len(old_list) > 0:
            self.junction_properties_list.delete(*old_list)

        # Get the new set of parameters
        prop_list = self.get_properties_list(item_id)

        # We scan the list of required properties and load the overwritten ones of the default ones
        for key in prop_list:
            try:
                prop = current_item['options'][key]
                self.junction_properties_list.insert('', 'end', text=key, values=(format(prop, '.4'), units[key]),
                                                     tags=('modified'))
            except:
                prop = default_junction_properties[key] / conversion[key]

                self.junction_properties_list.insert('', 'end', text=key, values=(format(prop, '.4'), units[key]))

    def show_layer_properties_frame(self, item_id):
        """

        :return: None
        """

        current_item = self.model['Solar cell'][item_id]

        # We bring the properties frame back, in case it was hidden
        self.layer_properties_frame.grid()

        # # And enable the input
        # self.enable_disable_layer_properties_frame('enable')

        self.name_var.set(current_item['name'])
        self.width_var.set(current_item['width'])
        self.selected_material.set(current_item['material'])

        # Checking if the material is ternary (or more) by counting capital letters.
        # If it is not, we ignore the composition
        if sum(1 for c in current_item['material'] if c.isupper()) > 2:
            try:
                x_label = current_item['options']['element']
            except:
                x_label = solcore.ParameterSystem().database.get(current_item['material'], 'x')
                current_item['options']['element'] = x_label

            self.composition_label_var.set(x_label)
            x = current_item['options']['x'] if 'x' in current_item['options'].keys() else 0
            self.composition_var.set(x)
        else:
            x_label = '-'
            x = 0
            self.composition_label_var.set('-')
            self.composition_var.set(0)

        # Finally, we populate the properties TreeView with the values for this layer,
        # starting by eliminating the current ones
        old_list = self.properties_list.get_children()
        if len(old_list) > 0:
            self.properties_list.delete(*old_list)

        # Get the new set of parameters
        prop_list = self.get_properties_list(item_id)

        Na = current_item['options']['Na'] if 'Na' in current_item['options'].keys() else 0.0
        Nd = current_item['options']['Nd'] if 'Nd' in current_item['options'].keys() else 0.0

        # Most properties are the default ones defined for that material, so we need to open that material
        solcore_mat = solcore.material(current_item['material'])(
            **{'T': self.model['Global']['T'], x_label: x, 'Na': Na, 'Nd': Nd})

        # We scan the list of required properties and load the overwritten ones of the default ones
        for key in prop_list:
            try:
                prop = current_item['options'][key]
                self.properties_list.insert('', 'end', text=key, values=(format(prop, '.4'), units[key]),
                                            tags=('modified'))
            except:
                try:
                    # We try the default property fo that material
                    prop = solcore_mat.__getattr__(key) / conversion[key]
                    self.properties_list.insert('', 'end', text=key, values=(format(prop, '.4'), units[key]))
                except:
                    # Or a default value.
                    prop = default_layer_properties[key] / conversion[key]
                    current_item['options'][key] = prop
                    self.properties_list.insert('', 'end', text=key, values=(format(prop, '.4'), units[key]),
                                                tags=('modified'))

    def get_properties_list(self, item_id):
        """ Return the properties that a layer needs to have. Which properties are relevant depend on where the layer is located, if its outside any juction, if the junction is a PDD one, a DA, etc.

        :param item_id: The id of the item we want the properties
        :return: The list of relevant properties
        """
        item = self.model['Solar cell'][item_id]
        parent_id = self.solar_cell_list.parent(item_id)
        try:
            parent = self.model['Solar cell'][parent_id]
            return properties_layers[parent['type']]
        except KeyError:
            if 'Layer' in item['type']:
                return []
            else:
                return properties_junctions[item['type']]

    def confirm_changes_to_layer(self, *args):
        """ This command confirm the changes made to the layer, validating the values and updating the TreeView and the Model

        :param args: Dummy
        :return: None
        """

        item_id = self.solar_cell_list.focus()

        if item_id == '':
            return

        name = self.name_var.get()
        mat = self.selected_material.get()
        width = self.width_var.get()
        width = max(width, 0)  # <-- In case width is negative
        self.width_var.set(width)

        self.model['Solar cell'][item_id]['name'] = name
        self.model['Solar cell'][item_id]['width'] = width
        self.model['Solar cell'][item_id]['material'] = mat

        # Checking if the material is ternary (or more) by counting capital letters.
        # If it is not, we ignore the composition
        if sum(1 for c in mat if c.isupper()) == 3:
            x = self.composition_var.get()
            x = min(max(x, 0), 1)  # <-- In case x is outside the 0-1 range
            self.composition_var.set(x)
            self.model['Solar cell'][item_id]['options']['x'] = x
            self.model['Solar cell'][item_id]['options']['element'] = self.composition_label_var.get()
        else:
            try:
                self.model['Solar cell'][item_id]['options'].pop('x')
                self.model['Solar cell'][item_id]['options'].pop('element')
            except:
                pass

        options = ''
        for key in self.model['Solar cell'][item_id]['options']:
            options += '{0}={1} '.format(key, self.model['Solar cell'][item_id]['options'][key])

        # We update the layers list
        self.solar_cell_list.item(item_id, text=name, values=['Layer', width, mat, options])

        # And the whole properties frame. Part of what is done here is redundant and can be improved a lot
        self.show_properties_frame()

    def update_composition_label(self, *args):
        """ Updates the composition label with the relevant material name

        :param args:
        :return:
        """
        # Checking if the material is ternary (or more) by counting capital letters.
        # If it is not, we ignore the compositionon
        mat = self.materials_box.get()
        if sum(1 for c in mat if c.isupper()) > 2:
            x_label = solcore.ParameterSystem().database.get(mat, 'x')
            self.composition_label_var.set(x_label)
        else:
            self.composition_label_var.set('-')

    def get_solar_cell_structure(self):
        """ Gets the structure of the solar cell from the tree view (order of the layers, junctions and layers within each junction).

        :return:
        """
        # First we get the structure of the solar cell, just using the labels. The first label in each list is the
        # parent
        structure = []
        children = self.solar_cell_list.get_children()
        for c in children:
            structure.append([c])
            grandchildren = self.solar_cell_list.get_children(c)
            for gc in grandchildren:
                structure[-1].append(gc)

        self.model['Solar cell']['structure'] = structure
        self.confirm_changes_to_solar_cell()

    def save_model(self):
        """ Saves the solar cell model to a Yaml-formatted file

        :return: None
        """

        self.get_solar_cell_structure()

        try:
            filename = filedialog.asksaveasfilename(defaultextension='.sol', initialdir=self.current_location,
                                                    initialfile=self.current_filename)
            with open(filename, 'w') as f:
                yaml.dump(self.model, f)

            self.current_location, self.current_filename = os.path.split(filename)
            self.wd_label_var.set(self.current_location)
        except:
            return 1

    def load_model(self):
        """ Loads a solar cell model previously saved in a Yalm-formatted file

        :return: None
        """

        # First we eliminate the existing elements of the solar cell list
        self.clear_solar_cell(confirmed=True)

        # We load the data from a Solcore file
        try:
            filename = filedialog.askopenfilename(initialdir=self.current_location,
                                                  defaultextension='.sol',
                                                  filetypes=[('All files', '*.*'), ('Solcore', '*.sol')],
                                                  initialfile=self.current_filename)
            with open(filename, 'r') as f:
                self.model = yaml.load(f)

            self.current_location, self.current_filename = os.path.split(filename)
            self.wd_label_var.set(self.current_location)
            self.refresh_plots_list()
        except:
            return

        try:
            # And now we populate the list with the loaded elements: junctions, tunnel junctions or layers
            for c in self.model['Solar cell']['structure']:
                item_id = c[0]
                current_item = self.model['Solar cell'][item_id]

                options = ''
                for key in current_item['options']:
                    options += '{0}={1} '.format(key, current_item['options'][key])

                self.solar_cell_list.insert('', tk.END, iid=item_id, text=current_item['name'], open=True,
                                            values=(
                                                current_item['type'], current_item['width'], current_item['material'],
                                                options))

                # Now we populate the layers within the junctions, if any
                for id in c[1:]:
                    layer = self.model['Solar cell'][id]

                    options = ''
                    for key in layer['options']:
                        options += '{0}={1} '.format(key, layer['options'][key])

                    self.solar_cell_list.insert(item_id, tk.END, iid=id, text=layer['name'], open=True,
                                                values=(layer['type'], layer['width'], layer['material'], options))
        except KeyError:
            print("The model cannot be loaded. There is missing information or the format is not correct.")

    def run(self, *args):
        """ Runs the selected calculation, starting by saving the solar cell model

        :param args:
        :return:
        """

        # We get the solar cell structure and if there is no structure, we do nothing else.
        self.get_solar_cell_structure()
        if len(self.model['Solar cell']['structure']) == 0:
            return

        # Get problem to solve
        list_of_tasks = ['optics', 'iv', 'qe', 'equilibrium', 'short_circuit']
        task = list_of_tasks[self.run_var.get() - 1]

        # Run model
        output = run_solar_cell_model(task, self.model)

        # Get output
        # And now we save the data to individual files for IV, QE and bandstructure (only for PDD junctions)
        # And plot the results
        # And show the calculated outputs, if any
        root_filename = os.path.join(self.current_location, self.current_filename.split('.')[0])
        if task == 'iv':
            light_iv = self.model['Electrical']['light_iv'] if 'light_iv' in self.model['Electrical'].keys() else False
            df, filename = save_iv_data(root_filename, output, light_iv)
            plot_name = '{} {}'.format(os.path.basename(filename.split('.')[0]), time.asctime())
            plot = IVPlot(df, light=light_iv, title=plot_name)
            save_plot(plot, filename.replace('csv', 'pkl'))
            self.show_plot_outputs(plot)

        elif task == 'qe':
            df, filename = save_qe_data(root_filename, output)
            plot_name = '{} {}'.format(os.path.basename(filename.split('.')[0]), time.asctime())
            plot = QEPlot(df, title=plot_name)
            save_plot(plot, filename.replace('csv', 'pkl'))
            self.show_plot_outputs(plot)

        elif task in ['equilibrium', 'short_circuit']:
            print('WARNING: Only data for PDD junctions will be saved.')
            df, filename = save_bandstructure_data(root_filename, output, task)
            plot_name = '{} {}'.format(os.path.basename(filename.split('.')[0]), time.asctime())
            plot = BandstructurePlot(df, title=plot_name)
            save_plot(plot, filename.replace('csv', 'pkl'))
            
        elif task == 'optics':
            pass

        self.refresh_plots_list()

    def create_output_frame(self):
        """

        :return:
        """

        # Frame containing buttons controlling the output and a few buttons
        out_frame = ttk.LabelFrame(self, text='Plot actions')
        out_frame.grid(column=0, row=100, rowspan=99, sticky=tk.NSEW)
        out_frame.columnconfigure(0, weight=1)

        choose_directory_button = ttk.Button(out_frame, text='Choose working directory', command=self.choose_directory)
        choose_directory_button.grid(column=0, row=0, sticky=tk.NSEW)
        refresh_plots_button = ttk.Button(out_frame, text='Refresh plots list', command=self.refresh_plots_list)
        refresh_plots_button.grid(column=0, row=1, sticky=tk.NSEW)

        # We create the list of figures available in the current directory
        self.figures_list = ttk.Treeview(self)
        figures_list_scroll = ttk.Scrollbar(self, orient=tk.VERTICAL, command=self.figures_list.yview)
        self.figures_list.configure(yscrollcommand=figures_list_scroll.set)
        self.figures_list.grid(column=2, row=100, columnspan=3, rowspan=99, sticky=tk.NSEW)
        figures_list_scroll.grid(column=1, row=100, sticky=tk.NS)

        # When selecting an item, we load the corresponding figure
        self.figures_list.bind("<Double-Button-1>", self.open_plot)
        self.figures_list.bind("<<TreeviewSelect>>", self.show_plot)

        self.figures_list['columns'] = ('type')
        self.figures_list.heading('#0', text='Filename', anchor='w')
        self.figures_list.column("#0", stretch=tk.YES)
        self.figures_list.heading('type', text='Plot type', anchor='w')
        self.figures_list.column("type", stretch=tk.YES)

        # And a list containing the calculated parameters, if any
        self.plot_outputs_list = ttk.Treeview(self)
        self.plot_outputs_list.grid(column=10, row=100, columnspan=2, sticky=(tk.NSEW))

        self.plot_outputs_list['columns'] = ('value', 'units')
        self.plot_outputs_list.heading('#0', text='Variable', anchor='w')
        self.plot_outputs_list.column("#0", minwidth=0, width=180, stretch=tk.NO)
        self.plot_outputs_list.heading('value', text='Value', anchor='w')
        self.plot_outputs_list.column("value", minwidth=0, width=100, stretch=tk.NO)
        self.plot_outputs_list.heading('units', text='Units', anchor='w')
        self.plot_outputs_list.column("units", minwidth=0, width=100, stretch=tk.NO)

    def show_plot(self, *args) -> None:
        """ Brings to the foregrund a figure already plotted

        :param args: dummy
        :return: None
        """
        id = self.figures_list.focus()

        if self.plot_outputs[id]['object'] is not None:
            raise_window(self.plot_outputs[id]['object'].fig.number)
            self.show_plot_outputs(self.plot_outputs[id]['object'])

    def open_plot(self, *args) -> None:
        """ Opens the selected plot. If it was already oppened, closes it - actually just destroys the figure - and opens it again.

        :return: None
        """
        id = self.figures_list.focus()

        if self.plot_outputs[id]['object'] is not None:
            plt.close(self.plot_outputs[id]['object'].fig.number)
            self.plot_outputs[id]['object'] = None

        self.plot_outputs[id]['object'] = load_plot(self.plot_outputs[id]['path'], plot=True)
        self.show_plot_outputs(self.plot_outputs[id]['object'])

    def show_plot_outputs(self, plot):
        """ Gets the output calculations of the plot, if any, and shows them

        :return:
        """
        if not hasattr(plot, 'outputs'):
            return

        # First we clear the existing list
        old_list = self.plot_outputs_list.get_children()
        if len(old_list) > 0:
            self.plot_outputs_list.delete(*old_list)

        for key in plot.outputs.keys():
            self.plot_outputs_list.insert('', tk.END, text=key, values=(format(plot.outputs[key], '.4'), ''))

    def refresh_plots_list(self) -> None:
        """

        :return:
        """
        # First we clear the existing list
        old_list = self.figures_list.get_children()
        if len(old_list) > 0:
            self.figures_list.delete(*old_list)

        # The we populate it with new data
        available_plots = glob.glob(os.path.join(self.current_location, '*.pkl'))
        self.plot_outputs = {}

        for f in available_plots:
            name = os.path.basename(f).split('.')[0]

            if 'QE' in name:
                fig_type = 'QE'
            elif 'IV' in name:
                fig_type = 'IV'
            elif any(a in name for a in ['equilibrium', 'short_circuit']):
                fig_type = 'Bandstructure'
            else:
                fig_type = 'Unknown'

            id = self.figures_list.insert('', tk.END, text=name, values=[fig_type])
            self.plot_outputs[id] = {'path': f, 'object': None}

    def choose_directory(self) -> None:
        """ Opens a pop-up window to choose the working directory

        :return: None
        """
        out = filedialog.askdirectory(initialdir=self.current_location)

        if out != '':
            self.current_location = out
            self.wd_label_var.set(self.current_location)
            self.refresh_plots_list()


def raise_window(figname=None):
    """
    Raise the plot window for Figure figname to the foreground.  If no argument
    is given, raise the current figure.

    This function will only work with a Tk graphics backend.  It assumes you
    have already executed the command 'import matplotlib.pyplot as plt'.

    Function taken from:
    http://physicalmodelingwithpython.blogspot.com/2015/07/raising-figure-window-to-foreground.html

    """

    if figname:
        plt.figure(figname)
    cfm = plt.get_current_fig_manager()
    cfm.window.attributes('-topmost', True)
    cfm.window.attributes('-topmost', False)
    return cfm



import matplotlib

matplotlib.use("TkAgg")

import tkinter as tk
from tkinter import ttk


# import numpy as np
import sys
import datetime

# import os
# from datetime import datetime


# from tkinter import filedialog, messagebox
# from tkinter import font

from sunglass.materials import MaterialsTab
from sunglass.solar_cells import SolarCellsTab
from sunglass.log import LogTab
from sunglass.spectrum import SpectrumTab


class Sunglass(tk.Tk):
    """This class creates the main Sunglass window, the one that will serve as entry point for any other tool"""

    def __init__(self):
        """Constructor of the class

        :return: None
        """

        tk.Tk.__init__(self)

        self.title("Sunglass")
        self.protocol(
            "WM_DELETE_WINDOW", self.__quit
        )  # Used to force a "safe closing" of the program
        self.resizable(False, False)
        self.option_add("*tearOff", False)  # Prevents tearing the menus

        # We create the global variables
        self.create_global_variables()

        # We create the GUI
        self.create_gui()

        # Finally, we initiate the main loop.
        # This is a hack found here: http://github.com/matplotlib/matplotlib/issues/9637
        # to avoid a crashing that happens when combining certain versions of tcl (what's behind tkinter) and certain
        # versions of python
        while self.closed is False:
            try:
                self.update_idletasks()
                self.update()
            except UnicodeDecodeError:
                print("Caught Scroll Error")

    def __quit(self):
        """Quit the program

        :return: None
        """
        self.closed = True

        # Needed to prevent weird errors related with Tcl/Tk. See:
        # https://stackoverflow.com/questions/15448914/python-tkinter-ttk-combobox-throws-exception-on-quit
        self.eval("::ttk::CancelRepeat")

        # And finally closeing the app
        self.destroy()  # destroys the main window
        self.quit()  # quits the program

    def create_global_variables(self):
        """Creates the global variables needed in several places of the program

        :return: None
        """
        self.closed = False

    def create_gui(self):
        """Creates the graphic components of the program, calling the relevant classes to add their bit (materials, solar cells, etc).

        :return: None
        """

        masterframe = ttk.Frame(self)
        masterframe.grid(column=0, row=0, sticky=tk.NSEW)
        masterframe.rowconfigure(99, weight=1)

        self.book = ttk.Notebook(masterframe)
        self.book.grid(column=0, row=0, sticky=tk.NSEW)

        # This needs to be called before the others
        self.materials = MaterialsTab(self.book, self)
        self.solar_cells = SolarCellsTab(self.book, self)
        self.output = LogTab(self.book)
        self.spectrum = SpectrumTab(self.book)

        # self.book.add(self.spectrum, text='Spectrum')
        self.book.add(self.solar_cells, text="Solar Cells")
        # self.book.add(self.materials, text='Materials')
        self.book.add(self.output, text="Log")

        sys.stdout = TextRedirector(self.output.text, "stdout")
        sys.stderr = TextRedirector(self.output.text, "stderr")


class TextRedirector(object):
    """Writes text in a chose widget. Taken from:

    https://stackoverflow.com/questions/12351786/how-to-redirect-print-statements-to-tkinter-text-widget
    """

    def __init__(self, widget, tag="stdout"):
        self.widget = widget
        self.tag = tag

    def write(self, str):
        if str == "\n":
            return
        now = datetime.datetime.now().isoformat(" ") + " "
        self.widget.configure(state="normal")
        self.widget.insert("end", now, (self.tag, "red"))
        self.widget.insert("end", str + "\n", (self.tag))
        self.widget.configure(state="disabled")


if __name__ == "__main__":
    window = Sunglass()
